@@ -15,64 +15,14 @@
  * along with this program; if not, see <https://www.gnu.org/licenses/>.
  */
 #include <lttoolbox/compiler.h>
-#include <lttoolbox/compression.h>
-#include <lttoolbox/entry_token.h>
-#include <lttoolbox/lt_locale.h>
 #include <lttoolbox/xml_parse_util.h>
 #include <lttoolbox/string_utils.h>
 #include <lttoolbox/file_utils.h>
+#include <lttoolbox/acx.h>
+#include <lttoolbox/regexp_compiler.h>
 
-#include <string>
-#include <cstdlib>
 #include <iostream>
-#include <libxml/encoding.h>
-
-using namespace std;
-
-UString const Compiler::COMPILER_DICTIONARY_ELEM    = "dictionary"_u;
-UString const Compiler::COMPILER_ALPHABET_ELEM      = "alphabet"_u;
-UString const Compiler::COMPILER_SDEFS_ELEM         = "sdefs"_u;
-UString const Compiler::COMPILER_SDEF_ELEM          = "sdef"_u;
-UString const Compiler::COMPILER_N_ATTR             = "n"_u;
-UString const Compiler::COMPILER_PARDEFS_ELEM       = "pardefs"_u;
-UString const Compiler::COMPILER_PARDEF_ELEM        = "pardef"_u;
-UString const Compiler::COMPILER_PAR_ELEM           = "par"_u;
-UString const Compiler::COMPILER_ENTRY_ELEM         = "e"_u;
-UString const Compiler::COMPILER_RESTRICTION_ATTR   = "r"_u;
-UString const Compiler::COMPILER_RESTRICTION_LR_VAL = "LR"_u;
-UString const Compiler::COMPILER_RESTRICTION_RL_VAL = "RL"_u;
-UString const Compiler::COMPILER_PAIR_ELEM          = "p"_u;
-UString const Compiler::COMPILER_LEFT_ELEM          = "l"_u;
-UString const Compiler::COMPILER_RIGHT_ELEM         = "r"_u;
-UString const Compiler::COMPILER_S_ELEM             = "s"_u;
-UString const Compiler::COMPILER_M_ELEM             = "m"_u;
-UString const Compiler::COMPILER_REGEXP_ELEM        = "re"_u;
-UString const Compiler::COMPILER_SECTION_ELEM       = "section"_u;
-UString const Compiler::COMPILER_ID_ATTR            = "id"_u;
-UString const Compiler::COMPILER_TYPE_ATTR          = "type"_u;
-UString const Compiler::COMPILER_IDENTITY_ELEM      = "i"_u;
-UString const Compiler::COMPILER_IDENTITYGROUP_ELEM = "ig"_u;
-UString const Compiler::COMPILER_JOIN_ELEM          = "j"_u;
-UString const Compiler::COMPILER_BLANK_ELEM         = "b"_u;
-UString const Compiler::COMPILER_POSTGENERATOR_ELEM = "a"_u;
-UString const Compiler::COMPILER_GROUP_ELEM         = "g"_u;
-UString const Compiler::COMPILER_LEMMA_ATTR         = "lm"_u;
-UString const Compiler::COMPILER_IGNORE_ATTR        = "i"_u;
-UString const Compiler::COMPILER_IGNORE_YES_VAL     = "yes"_u;
-UString const Compiler::COMPILER_ALT_ATTR           = "alt"_u;
-UString const Compiler::COMPILER_V_ATTR             = "v"_u;
-UString const Compiler::COMPILER_VL_ATTR            = "vl"_u;
-UString const Compiler::COMPILER_VR_ATTR            = "vr"_u;
-UString const Compiler::COMPILER_WEIGHT_ATTR        = "w"_u;
-UString const Compiler::COMPILER_TEXT_NODE          = "#text"_u;
-UString const Compiler::COMPILER_COMMENT_NODE       = "#comment"_u;
-UString const Compiler::COMPILER_ACX_ANALYSIS_ELEM  = "analysis-chars"_u;
-UString const Compiler::COMPILER_ACX_CHAR_ELEM      = "char"_u;
-UString const Compiler::COMPILER_ACX_EQUIV_CHAR_ELEM= "equiv-char"_u;
-UString const Compiler::COMPILER_ACX_VALUE_ATTR     = "value"_u;
-UString const Compiler::COMPILER_LSX_WB_ELEM        = "d"_u;
-UString const Compiler::COMPILER_LSX_CHAR_ELEM      = "w"_u;
-UString const Compiler::COMPILER_LSX_TAG_ELEM       = "t"_u;
+#include <thread>
 
 Compiler::Compiler()
 {
@@ -83,35 +33,24 @@ Compiler::~Compiler()
 }
 
 void
-Compiler::parseACX(string const &file, UString const &dir)
+Compiler::parseACX(std::string const &file, UStringView dir)
 {
   if(dir == COMPILER_RESTRICTION_LR_VAL)
   {
-    reader = xmlReaderForFile(file.c_str(), NULL, 0);
-    if(reader == NULL)
-    {
-      cerr << "Error: cannot open '" << file << "'." << endl;
-      exit(EXIT_FAILURE);
-    }
-    int ret = xmlTextReaderRead(reader);
-    while(ret == 1)
-    {
-      procNodeACX();
-      ret = xmlTextReaderRead(reader);
-    }
+    acx_map = readACX(file.c_str());
   }
 }
 
 void
-Compiler::parse(string const &file, UString const &dir)
+Compiler::parse(std::string const &file, UStringView dir)
 {
-  direction = dir;
-  reader = xmlReaderForFile(file.c_str(), NULL, 0);
-  if(reader == NULL)
-  {
-    cerr << "Error: Cannot open '" << file << "'." << endl;
-    exit(EXIT_FAILURE);
+  if (dir == COMPILER_RESTRICTION_U_VAL) {
+    direction = COMPILER_RESTRICTION_LR_VAL;
+    unified_compilation = true;
+  } else {
+    direction = dir;
   }
+  reader = XMLParseUtil::open_or_exit(file.c_str());
 
   int ret = xmlTextReaderRead(reader);
   while(ret == 1)
@@ -122,7 +61,7 @@ Compiler::parse(string const &file, UString const &dir)
 
   if(ret != 0)
   {
-    cerr << "Error: Parse error at the end of input." << endl;
+    std::cerr << "Error: Parse error at the end of input." << std::endl;
   }
 
   xmlFreeTextReader(reader);
@@ -133,7 +72,7 @@ Compiler::parse(string const &file, UString const &dir)
   // its own thread. This is the major bottleneck of lt-comp and sections
   // are completely independent transducers.
   std::vector<std::thread> minimisations;
-  for(std::pair<const UString, Transducer>& it : sections)
+  for(auto& it : sections)
   {
     if(jobs) {
       minimisations.push_back(
@@ -148,28 +87,42 @@ Compiler::parse(string const &file, UString const &dir)
     thr.join();
   }
 
+  if (is_separable) {
+    // ensure that all paths end in <$>, in case the user forgot to include
+    // <d/>. This will result in some paths ending with multiple finals
+    // and multiple finals, but lsx-proc only checks for finals upon reading
+    // $, so it won't be an issue.
+    int32_t end = alphabet(word_boundary, word_boundary);
+    for (auto& it : sections) {
+      for (auto fin : it.second.getFinals()) {
+        int end_state = it.second.insertSingleTransduction(end, fin.first);
+        it.second.setFinal(end_state);
+      }
+    }
+  }
+
   if (!valid(dir)) {
     exit(EXIT_FAILURE);
   }
 }
 
 bool
-Compiler::valid(UString const& dir) const
+Compiler::valid(UStringView dir) const
 {
-  const char* side = dir == COMPILER_RESTRICTION_RL_VAL ? "right" : "left";
-  const set<int> epsilonSymbols = alphabet.symbolsWhereLeftIs(0);
-  const set<int> spaceSymbols = alphabet.symbolsWhereLeftIs(' ');
+  const char* side = (dir == COMPILER_RESTRICTION_RL_VAL ? "right" : "left");
+  const std::set<int> epsilonSymbols = alphabet.symbolsWhereLeftIs(0);
+  const std::set<int> spaceSymbols = alphabet.symbolsWhereLeftIs(' ');
   for (auto &section : sections) {
     auto &fst = section.second;
     auto finals = fst.getFinals();
     auto initial = fst.getInitial();
     for(const auto i : fst.closure(initial, epsilonSymbols)) {
       if (finals.count(i)) {
-        cerr << "Error: Invalid dictionary (hint: the " << side << " side of an entry is empty)" << endl;
+        std::cerr << "Error: Invalid dictionary (hint: the " << side << " side of an entry is empty)" << std::endl;
         return false;
       }
       if(fst.closure(i, spaceSymbols).size() > 1) { // >1 since closure always includes self
-        cerr << "Error: Invalid dictionary (hint: entry on the " << side << " beginning with whitespace)" << endl;
+        std::cerr << "Error: Invalid dictionary (hint: entry on the " << side << " beginning with whitespace)" << std::endl;
         return false;
       }
     }
@@ -204,8 +157,8 @@ Compiler::procAlphabet()
     }
     else
     {
-      cerr << "Error (" << xmlTextReaderGetParserLineNumber(reader);
-      cerr << "): Missing alphabet symbols." << endl;
+      std::cerr << "Error (" << xmlTextReaderGetParserLineNumber(reader);
+      std::cerr << "): Missing alphabet symbols." << std::endl;
       exit(EXIT_FAILURE);
     }
   }
@@ -238,12 +191,12 @@ Compiler::procParDef()
 }
 
 int
-Compiler::matchTransduction(vector<int> const &pi,
-                           vector<int> const &pd,
+Compiler::matchTransduction(std::vector<int> const &pi,
+                           std::vector<int> const &pd,
                            int state, Transducer &t,
                            double const &entry_weight)
 {
-  vector<int>::const_iterator left, right, limleft, limright;
+  std::vector<int>::const_iterator left, right, limleft, limright;
 
   if(direction == COMPILER_RESTRICTION_LR_VAL)
   {
@@ -267,7 +220,7 @@ Compiler::matchTransduction(vector<int> const &pi,
   }
   else
   {
-    map<int, set<int> >::iterator acx_map_ptr;
+    std::map<int32_t, sorted_vector<int32_t> >::iterator acx_map_ptr;
     int rsymbol = 0;
 
     while(true)
@@ -314,11 +267,26 @@ Compiler::matchTransduction(vector<int> const &pi,
 
       int new_state = t.insertSingleTransduction(tag, state, weight_value);
 
+      if (is_separable) {
+        // loop-back symbols for <ANY_TAG> and <ANY_CHAR>
+        if (tag == alphabet(0, any_tag) || tag == alphabet(0, any_char)) {
+          // rl compilation of a badly written rule
+          // having an epsilon with wildcard output will produce
+          // garbage output -- see https://github.com/apertium/apertium-separable/issues/8
+          std::cerr << "Warning: Cannot insert <t/> from empty input. Ignoring. (You probably want to specify exact tags when deleting a word.)" << std::endl;
+        } else if (tag == alphabet(any_tag, any_tag) ||
+                   tag == alphabet(any_char, any_char) ||
+                   tag == alphabet(any_tag, 0) ||
+                   tag == alphabet(any_char, 0)) {
+          t.linkStates(new_state, new_state, tag);
+        }
+      }
+
       if(acx_map_ptr != acx_map.end())
       {
         for(auto& it : acx_map_ptr->second)
         {
-          t.linkStates(state, new_state, alphabet(it ,rsymbol), weight_value);
+          t.linkStates(state, new_state, alphabet(it, rsymbol), weight_value);
         }
       }
       state = new_state;
@@ -330,12 +298,12 @@ Compiler::matchTransduction(vector<int> const &pi,
 
 
 void
-Compiler::requireEmptyError(UString const &name)
+Compiler::requireEmptyError(UStringView name)
 {
   if(!xmlTextReaderIsEmptyElement(reader))
   {
-    cerr << "Error (" << xmlTextReaderGetParserLineNumber(reader);
-    cerr << "): Non-empty element '<" << name << ">' should be empty." << endl;
+    std::cerr << "Error (" << xmlTextReaderGetParserLineNumber(reader);
+    std::cerr << "): Non-empty element '<" << name << ">' should be empty." << std::endl;
     exit(EXIT_FAILURE);
   }
 }
@@ -343,19 +311,11 @@ Compiler::requireEmptyError(UString const &name)
 bool
 Compiler::allBlanks()
 {
-  bool flag = true;
-  UString text = XMLParseUtil::readValue(reader);
-
-  for(auto c : text)
-  {
-    flag = flag && u_isspace(c);
-  }
-
-  return flag;
+  return XMLParseUtil::allBlanks(reader);
 }
 
 void
-Compiler::readString(vector<int> &result, UString const &name)
+Compiler::readString(std::vector<int> &result, UStringView name)
 {
   if(name == COMPILER_TEXT_NODE)
   {
@@ -399,18 +359,37 @@ Compiler::readString(vector<int> &result, UString const &name)
 
     if(!alphabet.isSymbolDefined(symbol))
     {
-      cerr << "Error (" << xmlTextReaderGetParserLineNumber(reader);
-      cerr << "): Undefined symbol '" << symbol << "'." << endl;
+      std::cerr << "Error (" << xmlTextReaderGetParserLineNumber(reader);
+      std::cerr << "): Undefined symbol '" << symbol << "'." << std::endl;
       exit(EXIT_FAILURE);
     }
 
     result.push_back(alphabet(symbol));
   }
+  else if (is_separable && name == COMPILER_LSX_TAG_ELEM) {
+    requireEmptyError(name);
+    result.push_back(any_tag);
+  }
+  else if (is_separable && name == COMPILER_LSX_CHAR_ELEM) {
+    requireEmptyError(name);
+    result.push_back(any_char);
+  }
+  else if (is_separable && name == COMPILER_LSX_WB_ELEM) {
+    requireEmptyError(name);
+    UString mode = attrib(COMPILER_LSX_SPACE_ATTR);
+    if (mode == COMPILER_LSX_SPACE_YES_VAL) {
+      result.push_back(word_boundary_s);
+    } else if (mode == COMPILER_LSX_SPACE_NO_VAL) {
+      result.push_back(word_boundary_ns);
+    } else {
+      result.push_back(word_boundary);
+    }
+  }
   else
   {
-    cerr << "Error (" << xmlTextReaderGetParserLineNumber(reader);
-    cerr << "): Invalid specification of element '<" << name;
-    cerr << ">' in this context." << endl;
+    std::cerr << "Error (" << xmlTextReaderGetParserLineNumber(reader);
+    std::cerr << "): Invalid specification of element '<" << name;
+    std::cerr << ">' in this context." << std::endl;
     exit(EXIT_FAILURE);
   }
 }
@@ -424,8 +403,8 @@ Compiler::skipBlanks(UString &name)
     {
       if(!allBlanks())
       {
-        cerr << "Error (" << xmlTextReaderGetParserLineNumber(reader);
-        cerr << "): Invalid construction." << endl;
+        std::cerr << "Error (" << xmlTextReaderGetParserLineNumber(reader);
+        std::cerr << "): Invalid construction." << std::endl;
         exit(EXIT_FAILURE);
       }
     }
@@ -436,13 +415,7 @@ Compiler::skipBlanks(UString &name)
 }
 
 void
-Compiler::skip(UString &name, UString const &elem)
-{
-  skip(name, elem, true);
-}
-
-void
-Compiler::skip(UString &name, UString const &elem, bool open)
+Compiler::skip(UString &name, UStringView elem, bool open)
 {
   xmlTextReaderRead(reader);
   name = XMLParseUtil::readName(reader);
@@ -459,8 +432,8 @@ Compiler::skip(UString &name, UString const &elem, bool open)
     {
       if(!allBlanks())
       {
-        cerr << "Error (" << xmlTextReaderGetParserLineNumber(reader);
-        cerr << "): Invalid construction." << endl;
+        std::cerr << "Error (" << xmlTextReaderGetParserLineNumber(reader);
+        std::cerr << "): Invalid construction." << std::endl;
         exit(EXIT_FAILURE);
       }
     }
@@ -470,8 +443,8 @@ Compiler::skip(UString &name, UString const &elem, bool open)
 
   if(name != elem)
   {
-    cerr << "Error (" << xmlTextReaderGetParserLineNumber(reader);
-    cerr << "): Expected '<" << slash << elem << ">'." << endl;
+    std::cerr << "Error (" << xmlTextReaderGetParserLineNumber(reader);
+    std::cerr << "): Expected '<" << slash << elem << ">'." << std::endl;
     exit(EXIT_FAILURE);
   }
 }
@@ -479,7 +452,7 @@ Compiler::skip(UString &name, UString const &elem, bool open)
 EntryToken
 Compiler::procIdentity(double const entry_weight, bool ig)
 {
-  vector<int> both_sides;
+  std::vector<int> both_sides;
 
   if(!xmlTextReaderIsEmptyElement(reader))
   {
@@ -499,14 +472,14 @@ Compiler::procIdentity(double const entry_weight, bool ig)
 
   if(verbose && first_element && (both_sides.front() == (int)' '))
   {
-    cerr << "Error (" << xmlTextReaderGetParserLineNumber(reader);
-    cerr << "): Entry begins with space." << endl;
+    std::cerr << "Error (" << xmlTextReaderGetParserLineNumber(reader);
+    std::cerr << "): Entry begins with space." << std::endl;
   }
   first_element = false;
   EntryToken e;
   if(ig)
   {
-    vector<int> right;
+    std::vector<int> right;
     right.push_back(static_cast<int>('#'));
     right.insert(right.end(), both_sides.begin(), both_sides.end());
     e.setSingleTransduction(both_sides, right, entry_weight);
@@ -521,7 +494,7 @@ Compiler::procIdentity(double const entry_weight, bool ig)
 EntryToken
 Compiler::procTransduction(double const entry_weight)
 {
-  vector<int> lhs, rhs;
+  std::vector<int> lhs, rhs;
   UString name;
 
   skip(name, COMPILER_LEFT_ELEM);
@@ -543,8 +516,8 @@ Compiler::procTransduction(double const entry_weight)
 
   if(verbose && first_element && (lhs.front() == (int)' '))
   {
-    cerr << "Error (" << xmlTextReaderGetParserLineNumber(reader);
-    cerr << "): Entry begins with space." << endl;
+    std::cerr << "Error (" << xmlTextReaderGetParserLineNumber(reader);
+    std::cerr << "): Entry begins with space." << std::endl;
   }
   first_element = false;
 
@@ -573,7 +546,7 @@ Compiler::procTransduction(double const entry_weight)
 }
 
 UString
-Compiler::attrib(UString const &name)
+Compiler::attrib(UStringView name)
 {
   return XMLParseUtil::attrib(reader, name);
 }
@@ -587,15 +560,15 @@ Compiler::procPar()
 
   if(!current_paradigm.empty() && paradigm_name == current_paradigm)
   {
-    cerr << "Error (" << xmlTextReaderGetParserLineNumber(reader);
-    cerr << "): Paradigm refers to itself '" << paradigm_name << "'." <<endl;
+    std::cerr << "Error (" << xmlTextReaderGetParserLineNumber(reader);
+    std::cerr << "): Paradigm refers to itself '" << paradigm_name << "'." << std::endl;
     exit(EXIT_FAILURE);
   }
 
   if(paradigms.find(paradigm_name) == paradigms.end())
   {
-    cerr << "Error (" << xmlTextReaderGetParserLineNumber(reader);
-    cerr << "): Undefined paradigm '" << paradigm_name << "'." << endl;
+    std::cerr << "Error (" << xmlTextReaderGetParserLineNumber(reader);
+    std::cerr << "): Undefined paradigm '" << paradigm_name << "'." << std::endl;
     exit(EXIT_FAILURE);
   }
   e.setParadigm(paradigm_name);
@@ -603,7 +576,7 @@ Compiler::procPar()
 }
 
 void
-Compiler::insertEntryTokens(vector<EntryToken> const &elements)
+Compiler::insertEntryTokens(std::vector<EntryToken> const &elements)
 {
   if(!current_paradigm.empty())
   {
@@ -631,8 +604,8 @@ Compiler::insertEntryTokens(vector<EntryToken> const &elements)
       }
       else
       {
-        cerr << "Error (" << xmlTextReaderGetParserLineNumber(reader);
-        cerr << "): Invalid entry token." << endl;
+        std::cerr << "Error (" << xmlTextReaderGetParserLineNumber(reader);
+        std::cerr << "): Invalid entry token." << std::endl;
         exit(EXIT_FAILURE);
       }
     }
@@ -702,15 +675,14 @@ Compiler::insertEntryTokens(vector<EntryToken> const &elements)
 
 
 void
-Compiler::requireAttribute(UString const &value, UString const &attrname,
-                           UString const &elemname)
+Compiler::requireAttribute(UStringView value, UStringView attrname, UStringView elemname)
 {
   if(value.empty())
   {
-    cerr << "Error (" << xmlTextReaderGetParserLineNumber(reader);
-    cerr << "): '<" << elemname;
-    cerr << "' element must specify non-void '";
-    cerr << attrname << "' attribute." << endl;
+    std::cerr << "Error (" << xmlTextReaderGetParserLineNumber(reader);
+    std::cerr << "): '<" << elemname;
+    std::cerr << "' element must specify non-void '";
+    std::cerr << attrname << "' attribute." << std::endl;
     exit(EXIT_FAILURE);
   }
 }
@@ -723,8 +695,8 @@ Compiler::procSection()
 
   if(type != XML_READER_TYPE_END_ELEMENT)
   {
-    UString const &id = attrib(COMPILER_ID_ATTR);
-    UString const &type = attrib(COMPILER_TYPE_ATTR);
+    const auto& id = attrib(COMPILER_ID_ATTR);
+    const auto& type = attrib(COMPILER_TYPE_ATTR);
     requireAttribute(id, COMPILER_ID_ATTR, COMPILER_SECTION_ELEM);
     requireAttribute(type, COMPILER_TYPE_ATTR, COMPILER_SECTION_ELEM);
 
@@ -738,24 +710,116 @@ Compiler::procSection()
   }
 }
 
+bool
+Compiler::filterEntry(UStringView value, UStringView filter, bool keep_on_empty_filter)
+{
+  if (value.empty()) return true;
+  else if (keep_on_empty_filter && filter.empty()) return true;
+  auto ops = StringUtils::split(value, u" ");
+  for (auto& it : ops) {
+    if (it == filter) return true;
+  }
+  return false;
+}
+
+void
+Compiler::symbolFilters(UStringView value, UStringView prefix, std::vector<std::vector<int32_t>>& symbols)
+{
+  if (value.empty()) return;
+  std::vector<int32_t> syms;
+  for (auto& it : StringUtils::split(value, u" ")) {
+    if (it.empty()) continue;
+    UString tag;
+    tag += '<';
+    tag += prefix;
+    tag += ':';
+    tag += it;
+    tag += '>';
+    alphabet.includeSymbol(tag);
+    syms.push_back(alphabet(tag));
+  }
+  if (!syms.empty()) symbols.push_back(syms);
+}
+
 void
 Compiler::procEntry()
 {
-  UString attribute = this->attrib(COMPILER_RESTRICTION_ATTR);
-  UString ignore    = this->attrib(COMPILER_IGNORE_ATTR);
-  UString altval    = this->attrib(COMPILER_ALT_ATTR);
-  UString varval    = this->attrib(COMPILER_V_ATTR);
-  UString varl      = this->attrib(COMPILER_VL_ATTR);
-  UString varr      = this->attrib(COMPILER_VR_ATTR);
-  UString wsweight  = this->attrib(COMPILER_WEIGHT_ATTR);
+  UString attribute = attrib(COMPILER_RESTRICTION_ATTR);
+  UString ignore    = attrib(COMPILER_IGNORE_ATTR);
+  UString altval    = attrib(COMPILER_ALT_ATTR);
+  UString varval    = attrib(COMPILER_V_ATTR);
+  UString varl      = attrib(COMPILER_VL_ATTR);
+  UString varr      = attrib(COMPILER_VR_ATTR);
+  UString wsweight  = attrib(COMPILER_WEIGHT_ATTR);
+
+  std::vector<EntryToken> elements;
 
   // if entry is masked by a restriction of direction or an ignore mark
-  if((!attribute.empty() && attribute != direction)
+  if (unified_compilation && ignore != COMPILER_IGNORE_YES_VAL) {
+    std::vector<std::vector<int32_t>> symbols;
+    symbolFilters(attribute, u"r", symbols);
+    symbolFilters(altval, u"alt", symbols);
+    symbolFilters(varval, u"v", symbols);
+    symbolFilters(varl, u"vl", symbols);
+    symbolFilters(varr, u"vr", symbols);
+    if (!symbols.empty()) {
+      bool multi = false;
+      for (auto& it : symbols) {
+        if (it.size() > 1) {
+          multi = true;
+          break;
+        }
+      }
+      if (multi) {
+        UString parname = "--"_u;
+        parname += attribute;
+        parname += '-';
+        parname += altval;
+        parname += '-';
+        parname += varval;
+        parname += '-';
+        parname += varl;
+        parname += '-';
+        parname += varr;
+        if (paradigms.find(parname) == paradigms.end()) {
+          std::vector<int32_t> re;
+          for (auto& it : symbols) {
+            if (it.size() == 1) {
+              re.push_back(it[0]);
+            } else {
+              re.push_back(static_cast<int32_t>('['));
+              re.insert(re.end(), it.begin(), it.end());
+              re.push_back(static_cast<int32_t>(']'));
+            }
+          }
+          EntryToken e;
+          e.setRegexp(re);
+          std::vector<EntryToken> vec(1, e);
+          parname.swap(current_paradigm);
+          insertEntryTokens(vec);
+          parname.swap(current_paradigm);
+        }
+        EntryToken e;
+        e.setParadigm(parname);
+        elements.push_back(e);
+      }
+      else {
+        std::vector<int> syms;
+        for (auto& it : symbols) {
+          syms.push_back(it[0]);
+        }
+        EntryToken e;
+        e.setSingleTransduction(syms, syms);
+        elements.push_back(e);
+      }
+    }
+  }
+  else if((!attribute.empty() && attribute != direction)
    || ignore == COMPILER_IGNORE_YES_VAL
-   || (!altval.empty() && altval != alt)
-   || (direction == COMPILER_RESTRICTION_RL_VAL && !varval.empty() && varval != variant)
-   || (direction == COMPILER_RESTRICTION_RL_VAL && !varl.empty() && varl != variant_left)
-   || (direction == COMPILER_RESTRICTION_LR_VAL && !varr.empty() && varr != variant_right))
+   || !filterEntry(altval, alt, false)
+   || !filterEntry(varval, variant, true)
+   || (direction == COMPILER_RESTRICTION_RL_VAL && !filterEntry(varl, variant_left, false))
+   || (direction == COMPILER_RESTRICTION_LR_VAL && !filterEntry(varr, variant_right, false)))
   {
     // parse to the end of the entry
     UString name;
@@ -775,15 +839,38 @@ Compiler::procEntry()
     weight = StringUtils::stod(wsweight);
   }
 
-  vector<EntryToken> elements;
+  if (entry_debugging && current_paradigm.empty()) {
+    UString ln = "Line near "_u;
+    ln += StringUtils::itoa(xmlTextReaderGetParserLineNumber(reader));
+    // Note that this line number will usually be a little bit wrong.
+    // This function actually returns the current line of the *parser*
+    // which is probably several lines past the element we're currently
+    // looking at.
+    UString c = attrib(u"c");
+    if (!c.empty()) {
+      ln += ' ';
+      ln += c;
+    }
+    std::vector<int32_t> empty;
+    std::vector<int32_t> debug_syms;
+    ustring_to_vec32(ln, debug_syms);
+    if (is_separable) {
+      debug_syms.push_back(word_boundary_s);
+    } else {
+      debug_syms.push_back(static_cast<int32_t>(' '));
+    }
+    EntryToken e;
+    e.setSingleTransduction(empty, debug_syms);
+    elements.push_back(e);
+  }
 
   while(true)
   {
     int ret = xmlTextReaderRead(reader);
     if(ret != 1)
     {
-      cerr << "Error (" << xmlTextReaderGetParserLineNumber(reader);
-      cerr << "): Parse error." << endl;
+      std::cerr << "Error (" << xmlTextReaderGetParserLineNumber(reader);
+      std::cerr << "): Parse error." << std::endl;
       exit(EXIT_FAILURE);
     }
     UString name = XMLParseUtil::readName(reader);
@@ -817,16 +904,17 @@ Compiler::procEntry()
 
       // detection of the use of undefined paradigms
 
-      UString const &p = elements.rbegin()->paradigmName();
+      const auto& p = elements.rbegin()->paradigmName();
 
-      if(paradigms.find(p) == paradigms.end())
+      auto it = paradigms.find(p);
+      if(it == paradigms.end())
       {
-        cerr << "Error (" << xmlTextReaderGetParserLineNumber(reader);
-        cerr << "): Undefined paradigm '" << p << "'." <<endl;
+        std::cerr << "Error (" << xmlTextReaderGetParserLineNumber(reader);
+        std::cerr << "): Undefined paradigm '" << p << "'." << std::endl;
         exit(EXIT_FAILURE);
       }
       // discard entries with empty paradigms (by the directions, normally)
-      if(paradigms[p].isEmpty())
+      if(it->second.isEmpty())
       {
         while(name != COMPILER_ENTRY_ELEM || type != XML_READER_TYPE_END_ELEMENT)
         {
@@ -848,46 +936,14 @@ Compiler::procEntry()
     }
     else
     {
-      cerr << "Error (" << xmlTextReaderGetParserLineNumber(reader);
-      cerr << "): Invalid inclusion of '<" << name << ">' into '<" << COMPILER_ENTRY_ELEM;
-      cerr << ">'." << endl;
+      std::cerr << "Error (" << xmlTextReaderGetParserLineNumber(reader);
+      std::cerr << "): Invalid inclusion of '<" << name << ">' into '<" << COMPILER_ENTRY_ELEM;
+      std::cerr << ">'." << std::endl;
       exit(EXIT_FAILURE);
     }
   }
 }
 
-void
-Compiler::procNodeACX()
-{
-  UString name = XMLParseUtil::readName(reader);
-  if(name == COMPILER_TEXT_NODE)
-  {
-    /* ignore */
-  }
-  else if(name == COMPILER_ACX_ANALYSIS_ELEM)
-  {
-    /* ignore */
-  }
-  else if(name == COMPILER_ACX_CHAR_ELEM)
-  {
-    acx_current_char = static_cast<int>(attrib(COMPILER_ACX_VALUE_ATTR)[0]);
-  }
-  else if(name == COMPILER_ACX_EQUIV_CHAR_ELEM)
-  {
-    acx_map[acx_current_char].insert(static_cast<int>(attrib(COMPILER_ACX_VALUE_ATTR)[0]));
-  }
-  else if(name == COMPILER_COMMENT_NODE)
-  {
-    /* ignore */
-  }
-  else
-  {
-    cerr << "Error in ACX file (" << xmlTextReaderGetParserLineNumber(reader);
-    cerr << "): Invalid node '<" << name << ">'." << endl;
-    exit(EXIT_FAILURE);
-  }
-}
-
 void
 Compiler::procNode()
 {
@@ -901,7 +957,20 @@ Compiler::procNode()
   }
   else if(name == COMPILER_DICTIONARY_ELEM)
   {
-    /* ignore */
+    if (attrib(COMPILER_TYPE_ATTR) == COMPILER_SEPARABLE_VAL ||
+        attrib(COMPILER_TYPE_ATTR) == COMPILER_SEQUENTIAL_VAL) {
+      is_separable = true;
+      alphabet.includeSymbol(Transducer::ANY_TAG_SYMBOL);
+      alphabet.includeSymbol(Transducer::ANY_CHAR_SYMBOL);
+      alphabet.includeSymbol(Transducer::LSX_BOUNDARY_SYMBOL);
+      alphabet.includeSymbol(Transducer::LSX_BOUNDARY_SPACE_SYMBOL);
+      alphabet.includeSymbol(Transducer::LSX_BOUNDARY_NO_SPACE_SYMBOL);
+      any_tag          = alphabet(Transducer::ANY_TAG_SYMBOL);
+      any_char         = alphabet(Transducer::ANY_CHAR_SYMBOL);
+      word_boundary    = alphabet(Transducer::LSX_BOUNDARY_SYMBOL);
+      word_boundary_s  = alphabet(Transducer::LSX_BOUNDARY_SPACE_SYMBOL);
+      word_boundary_ns = alphabet(Transducer::LSX_BOUNDARY_NO_SPACE_SYMBOL);
+    }
   }
   else if(name == COMPILER_ALPHABET_ELEM)
   {
@@ -944,8 +1013,8 @@ Compiler::procNode()
   }
   else
   {
-    cerr << "Error (" << xmlTextReaderGetParserLineNumber(reader);
-    cerr << "): Invalid node '<" << name << ">'." << endl;
+    std::cerr << "Error (" << xmlTextReaderGetParserLineNumber(reader);
+    std::cerr << "): Invalid node '<" << name << ">'." << std::endl;
     exit(EXIT_FAILURE);
   }
 }
@@ -967,25 +1036,25 @@ Compiler::write(FILE *output)
 }
 
 void
-Compiler::setAltValue(UString const &a)
+Compiler::setAltValue(UStringView a)
 {
   alt = a;
 }
 
 void
-Compiler::setVariantValue(UString const &v)
+Compiler::setVariantValue(UStringView v)
 {
   variant = v;
 }
 
 void
-Compiler::setVariantLeftValue(UString const &v)
+Compiler::setVariantLeftValue(UStringView v)
 {
   variant_left = v;
 }
 
 void
-Compiler::setVariantRightValue(UString const &v)
+Compiler::setVariantRightValue(UStringView v)
 {
   variant_right = v;
 }
@@ -1013,3 +1082,9 @@ Compiler::setVerbose(bool verbosity)
 {
   verbose = verbosity;
 }
+
+void
+Compiler::setEntryDebugging(bool debug)
+{
+  entry_debugging = debug;
+}
