@@ -17,21 +17,15 @@
 
 #include <lttoolbox/compiler.h>
 #include <lttoolbox/expander.h>
-#include <lttoolbox/lt_locale.h>
 #include <lttoolbox/xml_parse_util.h>
-#include <lttoolbox/my_stdio.h>
 
 #include <cstdlib>
 #include <iostream>
 #include <libxml/encoding.h>
 
 
-using namespace std;
-
-Expander::Expander() :
-reader(0)
+Expander::Expander()
 {
-  LtLocale::tryToSetLocale();
 }
 
 Expander::~Expander()
@@ -39,14 +33,9 @@ Expander::~Expander()
 }
 
 void
-Expander::expand(string const &file, UFILE* output)
+Expander::expand(std::string const &file, UFILE* output)
 {
-  reader = xmlReaderForFile(file.c_str(), NULL, 0);
-  if(reader == NULL)
-  {
-    cerr << "Error: Cannot open '" << file << "'." << endl;
-    exit(EXIT_FAILURE);
-  }
+  reader = XMLParseUtil::open_or_exit(file.c_str());
 
   int ret = xmlTextReaderRead(reader);
   while(ret == 1)
@@ -57,7 +46,7 @@ Expander::expand(string const &file, UFILE* output)
 
   if(ret != 0)
   {
-    cerr << "Error: Parse error at the end of input." << endl;
+    std::cerr << "Error: Parse error at the end of input." << std::endl;
   }
 
   xmlFreeTextReader(reader);
@@ -80,12 +69,12 @@ Expander::procParDef()
 }
 
 void
-Expander::requireEmptyError(UString const &name)
+Expander::requireEmptyError(UStringView name)
 {
   if(!xmlTextReaderIsEmptyElement(reader))
   {
-    cerr << "Error (" << xmlTextReaderGetParserLineNumber(reader);
-    cerr << "): Non-empty element '<" << name << ">' should be empty." << endl;
+    std::cerr << "Error (" << xmlTextReaderGetParserLineNumber(reader);
+    std::cerr << "): Non-empty element '<" << name << ">' should be empty." << std::endl;
     exit(EXIT_FAILURE);
   }
 }
@@ -93,19 +82,11 @@ Expander::requireEmptyError(UString const &name)
 bool
 Expander::allBlanks()
 {
-  bool flag = true;
-  UString text = XMLParseUtil::readValue(reader);
-
-  for(auto c : text)
-  {
-    flag = flag && isspace(c);
-  }
-
-  return flag;
+  return XMLParseUtil::allBlanks(reader);
 }
 
 void
-Expander::readString(UString &result, UString const &name)
+Expander::readString(UString &result, UStringView name)
 {
   if(name == Compiler::COMPILER_TEXT_NODE)
   {
@@ -168,9 +149,9 @@ Expander::readString(UString &result, UString const &name)
   }
   else
   {
-    cerr << "Error (" << xmlTextReaderGetParserLineNumber(reader);
-    cerr << "): Invalid specification of element '<" << name;
-    cerr << ">' in this context." << endl;
+    std::cerr << "Error (" << xmlTextReaderGetParserLineNumber(reader);
+    std::cerr << "): Invalid specification of element '<" << name;
+    std::cerr << ">' in this context." << std::endl;
     exit(EXIT_FAILURE);
   }
 }
@@ -182,8 +163,8 @@ Expander::skipBlanks(UString &name)
   {
     if(!allBlanks())
     {
-      cerr << "Error (" << xmlTextReaderGetParserLineNumber(reader);
-      cerr << "): Invalid construction." << endl;
+      std::cerr << "Error (" << xmlTextReaderGetParserLineNumber(reader);
+      std::cerr << "): Invalid construction." << std::endl;
       exit(EXIT_FAILURE);
     }
     xmlTextReaderRead(reader);
@@ -192,7 +173,7 @@ Expander::skipBlanks(UString &name)
 }
 
 void
-Expander::skip(UString &name, UString const &elem)
+Expander::skip(UString &name, UStringView elem)
 {
   xmlTextReaderRead(reader);
   name = XMLParseUtil::readName(reader);
@@ -201,8 +182,8 @@ Expander::skip(UString &name, UString const &elem)
   {
     if(!allBlanks())
     {
-      cerr << "Error (" << xmlTextReaderGetParserLineNumber(reader);
-      cerr << "): Invalid construction." << endl;
+      std::cerr << "Error (" << xmlTextReaderGetParserLineNumber(reader);
+      std::cerr << "): Invalid construction." << std::endl;
       exit(EXIT_FAILURE);
     }
     xmlTextReaderRead(reader);
@@ -211,8 +192,8 @@ Expander::skip(UString &name, UString const &elem)
 
   if(name != elem)
   {
-    cerr << "Error (" << xmlTextReaderGetParserLineNumber(reader);
-    cerr << "): Expected '<" << elem << ">'." << endl;
+    std::cerr << "Error (" << xmlTextReaderGetParserLineNumber(reader);
+    std::cerr << "): Expected '<" << elem << ">'." << std::endl;
     exit(EXIT_FAILURE);
   }
 }
@@ -240,7 +221,7 @@ Expander::procIdentity()
   return both_sides;
 }
 
-pair<UString, UString>
+std::pair<UString, UString>
 Expander::procIdentityGroup()
 {
   UString lhs;
@@ -265,11 +246,11 @@ Expander::procIdentityGroup()
   lhs += both_sides;
   rhs += both_sides;
 
-  pair<UString, UString> e(lhs, rhs);
+  std::pair<UString, UString> e(lhs, rhs);
   return e;
 }
 
-pair<UString, UString>
+std::pair<UString, UString>
 Expander::procTransduction()
 {
   UString lhs, rhs;
@@ -311,12 +292,12 @@ Expander::procTransduction()
 
   skip(name, Compiler::COMPILER_PAIR_ELEM);
 
-  pair<UString, UString> e(lhs, rhs);
+  std::pair<UString, UString> e(lhs, rhs);
   return e;
 }
 
 UString
-Expander::attrib(UString const &name)
+Expander::attrib(UStringView name)
 {
   return XMLParseUtil::attrib(reader, name);
 }
@@ -330,15 +311,14 @@ Expander::procPar()
 }
 
 void
-Expander::requireAttribute(UString const &value, UString const &attrname,
-                           UString const &elemname)
+Expander::requireAttribute(UStringView value, UStringView attrname, UStringView elemname)
 {
   if(value.empty())
   {
-    cerr << "Error (" << xmlTextReaderGetParserLineNumber(reader);
-    cerr << "): '<" << elemname;
-    cerr << "' element must specify non-void '";
-    cerr<< attrname << "' attribute." << endl;
+    std::cerr << "Error (" << xmlTextReaderGetParserLineNumber(reader);
+    std::cerr << "): '<" << elemname;
+    std::cerr << "' element must specify non-void '";
+    std::cerr<< attrname << "' attribute." << std::endl;
     exit(EXIT_FAILURE);
   }
 }
@@ -367,8 +347,8 @@ Expander::procEntry(UFILE* output)
       int ret = xmlTextReaderRead(reader);
       if(ret != 1)
       {
-        cerr << "Error (" << xmlTextReaderGetParserLineNumber(reader);
-        cerr << "): Parse error." << endl;
+        std::cerr << "Error (" << xmlTextReaderGetParserLineNumber(reader);
+        std::cerr << "): Parse error." << std::endl;
         exit(EXIT_FAILURE);
       }
       myname = XMLParseUtil::readName(reader);
@@ -399,8 +379,8 @@ Expander::procEntry(UFILE* output)
     int ret = xmlTextReaderRead(reader);
     if(ret != 1)
     {
-      cerr << "Error (" << xmlTextReaderGetParserLineNumber(reader);
-      cerr << "): Parse error." << endl;
+      std::cerr << "Error (" << xmlTextReaderGetParserLineNumber(reader);
+      std::cerr << "): Parse error." << std::endl;
       exit(EXIT_FAILURE);
     }
     UString name = XMLParseUtil::readName(reader);
@@ -409,7 +389,7 @@ Expander::procEntry(UFILE* output)
     int type = xmlTextReaderNodeType(reader);
     if(name == Compiler::COMPILER_PAIR_ELEM)
     {
-      pair<UString, UString> p = procTransduction();
+      std::pair<UString, UString> p = procTransduction();
       append(items, p);
       append(items_lr, p);
       append(items_rl, p);
@@ -423,7 +403,7 @@ Expander::procEntry(UFILE* output)
     }
     else if(name == Compiler::COMPILER_IDENTITYGROUP_ELEM)
     {
-      pair<UString, UString> p = procIdentityGroup();
+      std::pair<UString, UString> p = procIdentityGroup();
       append(items, p);
       append(items_lr, p);
       append(items_rl, p);
@@ -444,8 +424,8 @@ Expander::procEntry(UFILE* output)
          paradigm_lr.find(p) == paradigm_lr.end() &&
          paradigm_rl.find(p) == paradigm_rl.end())
       {
-        cerr << "Error (" << xmlTextReaderGetParserLineNumber(reader);
-        cerr << "): Undefined paradigm '" << p << "'." <<endl;
+        std::cerr << "Error (" << xmlTextReaderGetParserLineNumber(reader);
+        std::cerr << "): Undefined paradigm '" << p << "'." << std::endl;
         exit(EXIT_FAILURE);
       }
 
@@ -526,9 +506,9 @@ Expander::procEntry(UFILE* output)
     }
     else
     {
-      cerr << "Error (" << xmlTextReaderGetParserLineNumber(reader);
-      cerr << "): Invalid inclusion of '<" << name << ">' into '<" << Compiler::COMPILER_ENTRY_ELEM;
-      cerr << ">'." << endl;
+      std::cerr << "Error (" << xmlTextReaderGetParserLineNumber(reader);
+      std::cerr << "): Invalid inclusion of '<" << name << ">' into '<" << Compiler::COMPILER_ENTRY_ELEM;
+      std::cerr << ">'." << std::endl;
       exit(EXIT_FAILURE);
     }
   }
@@ -539,7 +519,7 @@ Expander::procNode(UFILE *output)
 {
   UString name = XMLParseUtil::readName(reader);
 
-  // DO: optimize the execution order of this string "ifs"
+  // TODO: optimize the execution order of this string "ifs"
 
   if(name == Compiler::COMPILER_TEXT_NODE)
   {
@@ -583,8 +563,8 @@ Expander::procNode(UFILE *output)
   }
   else
   {
-    cerr << "Error (" << xmlTextReaderGetParserLineNumber(reader);
-    cerr << "): Invalid node '<" << name << ">'." << endl;
+    std::cerr << "Error (" << xmlTextReaderGetParserLineNumber(reader);
+    std::cerr << "): Invalid node '<" << name << ">'." << std::endl;
     exit(EXIT_FAILURE);
   }
 }
@@ -593,9 +573,25 @@ UString
 Expander::procRegexp()
 {
   xmlTextReaderRead(reader);
-  UString re = XMLParseUtil::readValue(reader);
+  UString val = XMLParseUtil::readValue(reader);
+  UString escaped = "^$/<>{}*@#+~:"_u;
+  UString ret;
+  bool esc = false;
+  for (auto& c : val) {
+    if (esc) {
+      ret += c;
+      esc = false;
+      continue;
+    }
+    if (escaped.find(c) != UString::npos) {
+      ret += '\\';
+    } else if (c == '\\') {
+      esc = true;
+    }
+    ret += c;
+  }
   xmlTextReaderRead(reader);
-  return re;
+  return ret;
 }
 
 void
@@ -608,7 +604,7 @@ Expander::append(EntList &result,
   {
     for(auto& it2 : endings)
     {
-      temp.push_back(pair<UString, UString>(it.first + it2.first,
+      temp.push_back(std::pair<UString, UString>(it.first + it2.first,
                           it.second + it2.second));
     }
   }
@@ -617,7 +613,7 @@ Expander::append(EntList &result,
 }
 
 void
-Expander::append(EntList &result, UString const &endings)
+Expander::append(EntList &result, UStringView endings)
 {
   for(auto& it : result)
   {
@@ -628,7 +624,7 @@ Expander::append(EntList &result, UString const &endings)
 
 void
 Expander::append(EntList &result,
-                 pair<UString, UString> const &endings)
+                 std::pair<UString, UString> const &endings)
 {
   for(auto& it : result)
   {
@@ -638,25 +634,25 @@ Expander::append(EntList &result,
 }
 
 void
-Expander::setAltValue(UString const &a)
+Expander::setAltValue(UStringView a)
 {
   alt = a;
 }
 
 void
-Expander::setVariantValue(UString const &v)
+Expander::setVariantValue(UStringView v)
 {
   variant = v;
 }
 
 void
-Expander::setVariantLeftValue(UString const &v)
+Expander::setVariantLeftValue(UStringView v)
 {
   variant_left = v;
 }
 
 void
-Expander::setVariantRightValue(UString const &v)
+Expander::setVariantRightValue(UStringView v)
 {
   variant_right = v;
 }
