@@ -15,71 +15,34 @@
  * along with this program; if not, see <https://www.gnu.org/licenses/>.
  */
 #include <lttoolbox/fst_processor.h>
-#include <lttoolbox/my_stdio.h>
 #include <lttoolbox/lt_locale.h>
-
-#include <cstdlib>
-#include <getopt.h>
-#include <iostream>
-#include <libgen.h>
-
-using namespace std;
-
-[[noreturn]]
-void endProgram(char *name)
-{
-  cout << basename(name) << ": process a stream with a letter transducer" << endl;
-  cout << "USAGE: " << basename(name) << " fst_file [input_file [output_file]]" << endl;
-  exit(EXIT_FAILURE);
-}
-
-void checkValidity(FSTProcessor const &fstp)
-{
-  if(!fstp.valid())
-  {
-    exit(EXIT_FAILURE);
-  }
-}
+#include <lttoolbox/cli.h>
+#include <lttoolbox/file_utils.h>
 
 int main(int argc, char *argv[])
 {
   LtLocale::tryToSetLocale();
+  CLI cli("process a stream with a letter transducer");
+  cli.add_file_arg("fst_file", false);
+  cli.add_file_arg("input_file");
+  cli.add_file_arg("output_file");
+  cli.parse_args(argc, argv);
 
-  InputFile input;
-  UFILE* output = u_finit(stdout, NULL, NULL);
   FSTProcessor fstp;
-  FILE *aux;
-
-  switch(argc)
-  {
-    case 4:
-      output = u_fopen(argv[3], "wb", NULL, NULL);
-      if(!output)
-      {
-        endProgram(argv[0]);
-      }
-      // follow
-    case 3:
-      if (!input.open(argv[2])) {
-        endProgram(argv[0]);
-      }
-      // follow
-    case 2:
-      aux = fopen(argv[1], "rb");
-      if(!aux)
-      {
-        endProgram(argv[0]);
-      }
-      fstp.load(aux);
-      fclose(aux);
-      break;
-    default:
-      endProgram(argv[0]);
-      break;
+  FILE* aux = openInBinFile(cli.get_files()[0]);
+  fstp.load(aux);
+  fclose(aux);
+  fstp.initTMAnalysis();
+  if (!fstp.valid()) {
+    return EXIT_FAILURE;
   }
 
-  fstp.initTMAnalysis();
-  checkValidity(fstp);
+  InputFile input;
+  if (!cli.get_files()[1].empty()) {
+    input.open_or_exit(cli.get_files()[1].c_str());
+  }
+  UFILE* output = openOutTextFile(cli.get_files()[2].c_str());
+
   fstp.tm_analysis(input, output);
 
   u_fclose(output);
