@@ -20,40 +20,43 @@
 #include <iostream>
 #include <utf8.h>
 
-using namespace std;
-
-UString
-XMLParseUtil::attrib(xmlTextReaderPtr reader, UString const &name)
+xmlTextReaderPtr
+XMLParseUtil::open_or_exit(const char* fname)
 {
-  return attrib(reader, name, ""_u);
+  xmlTextReaderPtr reader = xmlReaderForFile(fname, NULL, 0);
+  if (reader == NULL) {
+    std::cerr << "Error: cannot open '" << fname << "' for reading." << std::endl;
+    exit(EXIT_FAILURE);
+  }
+  return reader;
 }
 
 UString
-XMLParseUtil::attrib(xmlTextReaderPtr reader, UString const& name, const UString& fallback)
+XMLParseUtil::attrib(xmlTextReaderPtr reader, UStringView name, UStringView fallback)
 {
   std::string temp;
   temp.reserve(name.size());
   utf8::utf16to8(name.begin(), name.end(), std::back_inserter(temp));
-  const xmlChar *attrname = reinterpret_cast<const xmlChar*>(temp.c_str());
-  xmlChar *myattr = xmlTextReaderGetAttribute(reader, attrname);
+  auto attrname = reinterpret_cast<const xmlChar*>(temp.c_str());
+  auto myattr = xmlTextReaderGetAttribute(reader, attrname);
   if(myattr == NULL) {
     xmlFree(myattr);
-    return fallback;
+    return US(fallback);
   } else {
-    UString result = to_ustring(reinterpret_cast<char*>(myattr));
+    auto result = to_ustring(reinterpret_cast<char*>(myattr));
     xmlFree(myattr);
     return result;
   }
 }
 
 std::string
-XMLParseUtil::attrib_str(xmlTextReaderPtr reader, const UString& name)
+XMLParseUtil::attrib_str(xmlTextReaderPtr reader, UStringView name)
 {
   std::string temp;
   temp.reserve(name.size());
   utf8::utf16to8(name.begin(), name.end(), std::back_inserter(temp));
-  const xmlChar *attrname = reinterpret_cast<const xmlChar*>(temp.c_str());
-  xmlChar *myattr = xmlTextReaderGetAttribute(reader, attrname);
+  auto attrname = reinterpret_cast<const xmlChar*>(temp.c_str());
+  auto myattr = xmlTextReaderGetAttribute(reader, attrname);
   if(myattr == NULL) {
     xmlFree(myattr);
     return "";
@@ -81,7 +84,7 @@ XMLParseUtil::readValue(xmlTextReaderPtr reader)
 }
 
 void
-XMLParseUtil::readValueInto32(xmlTextReaderPtr reader, vector<int32_t>& vec)
+XMLParseUtil::readValueInto32(xmlTextReaderPtr reader, std::vector<int32_t>& vec)
 {
   const xmlChar* val = xmlTextReaderConstValue(reader);
   if (val == NULL) return;
@@ -89,3 +92,12 @@ XMLParseUtil::readValueInto32(xmlTextReaderPtr reader, vector<int32_t>& vec)
   vec.reserve(vec.size() + sz);
   utf8::utf8to32(val, val+sz, std::back_inserter(vec));
 }
+
+bool
+XMLParseUtil::allBlanks(xmlTextReaderPtr reader)
+{
+  for (auto& c : readValue(reader)) {
+    if (!u_isspace(c)) return false;
+  }
+  return true;
+}
